__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import io
from typing import Dict, List

import numpy as np

from .. import BaseSegmenter


class ImageReader(BaseSegmenter):
    """
    :class:`ImageReader` loads the image from the given file path and save the `ndarray` of the image in the Chunk.
    """

    def __init__(self, channel_axis: int = -1, from_bytes: bool = False, *args, **kwargs):
        """
        :class:`ImageReader` load an image file and craft into image matrix.

        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param from_bytes: if set to true, then load image directly from raw_bytes, otherwise it assumes raw_bytes as file path
        """
        super().__init__(*args, **kwargs)
        self.channel_axis = channel_axis
        self.from_bytes = from_bytes

    def craft(self, raw_bytes: bytes, doc_id: int, *args, **kwargs) -> List[Dict]:
        """
        Read the image from the given file path that specified in `raw_bytes` and save the `ndarray` of the image in
            the `blob` of the chunk.

        :param raw_bytes: the image file path in raw bytes
        :param doc_id: the id of the Document

        """
        from PIL import Image
        if self.from_bytes:
            raw_img = Image.open(io.BytesIO(raw_bytes))
        else:
            raw_img = Image.open(raw_bytes.decode())
        if raw_img.mode != 'RGB':
            raw_img = raw_img.convert('RGB')
        img = np.array(raw_img).astype('float32')
        if self.channel_axis != -1:
            img = np.moveaxis(img, -1, self.channel_axis)
        return [dict(doc_id=doc_id, offset=0, weight=1., blob=img), ]



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple

import numpy as np

from .numpy import NumpyIndexer


class AnnoyIndexer(NumpyIndexer):
    """Annoy powered vector indexer

    For more information about the Annoy supported parameters, please consult:
        - https://github.com/spotify/annoy

    .. note::
        Annoy package dependency is only required at the query time.
    """

    def __init__(self, metric: str = 'euclidean', n_trees: int = 10, *args, **kwargs):
        """
        Initialize an AnnoyIndexer

        :param metric: Metric can be "angular", "euclidean", "manhattan", "hamming", or "dot"
        :param n_trees: builds a forest of n_trees trees. More trees gives higher precision when querying.
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.metric = metric
        self.n_trees = n_trees

    def get_query_handler(self):
        vecs = super().get_query_handler()
        if vecs is not None:
            from annoy import AnnoyIndex
            _index = AnnoyIndex(self.num_dim, self.metric)
            vecs = vecs.astype(np.float32)
            for idx, v in enumerate(vecs):
                _index.add_item(idx, v)
            _index.build(self.n_trees)
            return _index
        else:
            return None

    def query(self, keys: 'np.ndarray', top_k: int, *args, **kwargs) -> Tuple['np.ndarray', 'np.ndarray']:
        if keys.dtype != np.float32:
            raise ValueError('vectors should be ndarray of float32')
        all_idx = []
        all_dist = []
        for k in keys:
            ret, dist = self.query_handler.get_nns_by_vector(k, top_k, include_distances=True)
            all_idx.append(self.int2ext_key[ret])
            all_dist.append(dist)
        return np.array(all_idx), np.array(all_dist)
